﻿using GridViewSamples.DataModel;
using System;
using System.Collections.Generic;
using System.IO;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Navigation;

// The Basic Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234237

namespace GridViewSamples.Samples
{
    /// <summary>
    /// This page shows customized GridViewEx control bound to the grouped data source. 
    /// Drag&drop is implemented in the GridViewEx control.
    /// Page state is persisted at navigating from this page. 
    /// The SaveState method performs the simplest serialization
    /// of the whole _groups collection to the JSON string and adds it to the page state.
    /// The LoadState method performs deserialization of the previously saved groups or creates groups from scratch if page state is empty.
    /// GridViewSamples.App class serializes this state along with application state using GridViewSamples.Common.SuspensionManager
    /// class. For more details see GridViewSamples.App code and code comments.
    /// Note, depending on your data structure and size you can implement persistence in the other way or save your data to some other place.
    /// If you need more information about saving the app's state, start from this article: 
    /// http://msdn.microsoft.com/en-us/library/windows/apps/hh986968.aspx.
    /// </summary>
    public sealed partial class Customized : GridViewSamples.Common.LayoutAwarePage
    {
        public double TileSize
        {
            get { return (double)GetValue(TileSizeProperty); }
            set { SetValue(TileSizeProperty, value); }
        }

        public static readonly DependencyProperty TileSizeProperty =
            DependencyProperty.Register(nameof(TileSize), typeof(double), typeof(Customized), new PropertyMetadata(100));

        List<Group> _groups = new List<Group>();

        public Customized()
        {
            if (IsWindowsPhoneDevice())
            {
                TileSize = 72;
            }
            this.InitializeComponent();
        }

        /// <summary>
        /// Populates the page with content passed during navigation.  Any saved state is also
        /// provided when recreating a page from a prior session.
        /// </summary>
        /// <param name="navigationParameter">The parameter value passed to
        /// <see cref="Frame.Navigate(Type, Object)"/> when this page was initially requested.
        /// </param>
        /// <param name="pageState">A dictionary of state preserved by this page during an earlier
        /// session.  This will be null the first time a page is visited.</param>
        protected override void LoadState(Object navigationParameter, Dictionary<String, Object> pageState)
        {
            base.LoadState(navigationParameter, pageState);
            if (pageState != null && pageState.Count > 0 && pageState.ContainsKey("Groups"))
            {
                // restore groups and items from the previously serialized state
                System.Runtime.Serialization.Json.DataContractJsonSerializer rootSer = 
                new System.Runtime.Serialization.Json.DataContractJsonSerializer(typeof(List<Group>));

                var stream = new MemoryStream(System.Text.Encoding.UTF8.GetBytes((string)pageState["Groups"]));
                _groups = (List<Group>)rootSer.ReadObject(stream);

            }
            else
            {
                // if we get here for the first time and don't have serialized content, fill groups and items from scratch
                for (int j = 1; j <= 12; j++)
                {
                    Group group = Group.GetNewGroup();

                    for (int i = 1; i <= 7 + j % 3; i++)
                    {
                        group.Items.Add(new Item()
                        {
                            Id = i,
                            GroupId = group.Id
                        });
                    }
                    _groups.Add(group);
                }
            }

            UpdateDataContext();
        }

        /// <summary>
        /// Preserves state associated with this page in case the application is suspended or the
        /// page is discarded from the navigation cache.  Values must conform to the serialization
        /// requirements of <see cref="SuspensionManager.SessionState"/>.
        /// </summary>
        /// <param name="pageState">An empty dictionary to be populated with serializable state.</param>
        protected override void SaveState(Dictionary<String, Object> pageState)
        {
            // save groups and items to JSON string so that it's possible to restore page state later
            base.SaveState(pageState);

            System.Runtime.Serialization.Json.DataContractJsonSerializer rootSer = 
                new System.Runtime.Serialization.Json.DataContractJsonSerializer(typeof(List<Group>));
            var stream = new MemoryStream();
            rootSer.WriteObject(stream, _groups);
            string str = System.Text.Encoding.UTF8.GetString(stream.ToArray(), 0, (int)stream.Length);
            pageState.Add("Groups", str);
        }

        /// <summary>
        /// Invoked when this page is about to be displayed in a Frame.
        /// </summary>
        /// <param name="e">Event data that describes how this page was reached.  The Parameter
        /// property is typically used to configure the page.</param>
        protected override void OnNavigatedTo(NavigationEventArgs e)
        {
            // restore page state
            var frameState = GridViewSamples.Common.SuspensionManager.SessionStateForFrame(this.Frame);
            if (frameState.ContainsKey("TilePageData"))
            {
                this.LoadState(e.Parameter, (Dictionary<String, Object>)frameState["TilePageData"]);
            }
            else
            {
                this.LoadState(e.Parameter, null);
            }
            base.OnNavigatedTo(e);
        }

        protected override void OnNavigatedFrom(NavigationEventArgs e)
        {
            // save page state with "TilePageData" key
            var frameState = GridViewSamples.Common.SuspensionManager.SessionStateForFrame(this.Frame);
            var pageState = new Dictionary<String, Object>();
            this.SaveState(pageState);
            frameState["TilePageData"] = pageState;
        }

        /// <summary>
        /// Creates new CollectionViewSource and updates page DataContext.
        /// </summary>
        private void UpdateDataContext()
        {
            CollectionViewSource source = new CollectionViewSource();
            source.Source = _groups;
            source.ItemsPath = new PropertyPath("Items");
            source.IsSourceGrouped = true;
            this.DataContext = source;
        }

        // creates new group in the data source, if end-user drags item to the new group placeholder
        private void gve_BeforeDrop(object sender, GridViewEx.BeforeDropItemsEventArgs e)
        {
            if (e.RequestCreateNewGroup)
            {
                // create new group and re-assign datasource 
                Group group = Group.GetNewGroup();
                _groups.Insert(e.NewGroupIndex, group);
                UpdateDataContext();
            }
        }

        // removes empty groups (except the last one)
        private void gve_Drop(object sender, DragEventArgs e)
        {
            bool needReset = false;
            for (int i = _groups.Count - 1; i >= 0; i--)
            {
                if (_groups[i].Items.Count == 0 && _groups.Count > 1)
                {
                    _groups.RemoveAt(i);
                    needReset = true;
                }
            }
            if (needReset)
            {
                UpdateDataContext();
            }
        }

        /// <summary>
        /// Set column spans depending on group id.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void gve_PreparingContainerForItem(object sender, GridViewEx.PreparingContainerForItemEventArgs e)
        {
            try
            {
                Item it = e.Item as Item;
                if (it != null)
                {
                    e.Element.SetValue(Windows.UI.Xaml.Controls.VariableSizedWrapGrid.ColumnSpanProperty, it.GroupId % 2 + 1);
                }
            }
            catch
            {
                e.Element.SetValue(Windows.UI.Xaml.Controls.VariableSizedWrapGrid.ColumnSpanProperty, 1);
            }
        }
    }
}
